/* eslint-disable camelcase */
import { randomInt } from '../../lib/utils/index.mjs';

/**
 * @function
 * @param {String} mac - hardware device MAC address
 * @param {String} sublocationUid - sublocation unique identifier
 * @param {String} tkUid - trackable unique id
 * @return {Object}
 */
export function createPosition(mac, sublocationUid, tkUid) {
    return {
        data: {
            fom: randomInt() % 100,
            mac,
            sl_uid: sublocationUid,
            type: 'TST',
            x: randomInt(),
            y: randomInt(),
            z: randomInt(),
        },
        tk_uid: tkUid,
        ts: new Date().toISOString(),
    };
}

/**
 *
 * @param {String} mac - hardware device MAC address
 * @param {String} sublocation_uid - sublocation unique identifier
 * @param {String} tk_uid - trackable unique id
 * @returns {{plist: {sl_id, ntype: string}, sublocation_uid, ms: number, x: number, action: string, y: number, z: number, mac, ts: number, dev_id, object: string}}
 */
export function createHeartbeat(mac, sublocation_uid, tk_uid) {
    const date = new Date();
    const ts = Math.floor(date.getTime() / 1000);
    const ms = date.getTime() % 1000;
    return {
        action: 'HB',
        dev_id: tk_uid,
        mac,
        ms,
        object: 'TO',
        plist: {
            ntype: 'NAV',
            sl_id: sublocation_uid,
        },
        sublocation_uid,
        ts,
        x: randomInt(),
        y: randomInt(),
        z: randomInt(),
    };
}

/**
 *
 * @param {Number} limit - position count to send
 * @param {String} mac - tag mac address
 * @param {String} sublocation_uid - sublocation unique id
 * @param {String} tk_uid - trackable unique id
 * @return {{next: (function(): {value: (undefined|{data:
 *      {fom: number, sl_uid: String, x: number, y: number, z: number, type: string, mac: String}, ts: string}), done: boolean}
 *   )}}
 * @constructor
 */
function PositionIterator({
    limit, mac, sublocation_uid, tk_uid,
}) {
    let index = 0;
    let done = false;
    return {
        // return the next method, need for iterator
        next: () => {
            done = (index >= limit);
            const obj = {
                done,
                value: done ? undefined : createPosition(mac, sublocation_uid, tk_uid),
            };
            index += 1; // increment index
            return obj;
        },
    };
}

/**
 *
 * @param param
 * @property {String} mac - hardware device MAC address
 * @property {Number} limit - heartbeat count
 * @property {String} sublocation_uid - sublocation unique identifier
 * @property {String} tk_uid - trackable unique id
 * @returns {{next: (function(): {value: (undefined|{plist: {sl_id, ntype: string},
 *      sublocation_uid, ms: number, x: number, action: string, y: number, z: number,
 *      mac, ts: number, dev_id, object: string}), done: boolean})}}
 * @constructor
 */
function HeartbeatIterator({
    limit, mac, sublocation_uid, tk_uid,
}) {
    let index = 0;
    let done = false;
    return {
        // return the next method, need for iterator
        next: () => {
            done = (index >= limit);
            const obj = {
                done,
                value: done ? undefined : createHeartbeat(mac, sublocation_uid, tk_uid),
            };
            index += 1; // increment index
            return obj;
        },
    };
}


/**
 *
 * @param param
 * @property {String} mac - hardware device MAC address
 * @property {Number} limit - position count
 * @property {String} sublocation_uid - sublocation unique identifier
 * @property {String} tk_uid - trackable unique id
 * @returns {{[Symbol.iterator]: (function():
 *      {next: (function(): {value: (undefined|{data: {fom: number, sl_uid: String, x: number, y: number,
 *      z: number, type: string, mac: String}, ts: string}), done: boolean})})}}
 */
export const makePositionsIterator = ({
    limit, mac, sublocation_uid, tk_uid,
}) => ({
    [Symbol.iterator]: () => PositionIterator({
        limit, mac, sublocation_uid, tk_uid,
    }),
});

/**
 *
 * @param param
 * @property {String} mac - hardware device MAC address
 * @property {Number} limit - heartbeat count
 * @property {String} sublocation_uid - sublocation unique identifier
 * @property {String} tk_uid - trackable unique id
 * @returns {{[Symbol.iterator]: (function(): {next: function():
 *  {value: (undefined|{plist: {sl_id, ntype: string}, sublocation_uid, ms: number, x: number, action: string,
 *  y: number, z: number, mac, ts: number, dev_id, object: string}), done: boolean}})}}
 */
export const makeHeartbeatIterator = (param) => ({
    [Symbol.iterator]: () => HeartbeatIterator(param),
});

export async function delay(ms) {
    return new Promise((resolve) => setTimeout(resolve, ms));
}
