import { createRepository } from '@rtls-platform/repository';
import { createLogger, LoggerTypes } from '@rtls-platform/logger';
import * as App from './lib/api/index.mjs';
import * as ZmqApi from './lib/api/zmq-api/app.mjs';
import * as WsApi from './lib/api/ws-api/app.mjs';
import * as ConfigContainer from './lib/config.cjs';
import { createRedisConnection, createMariaDbConnection } from './lib/db/index.mjs';
import ModelBase from './lib/models/ModelBase.mjs';
import { SubscriptionsManagerBase, State } from './lib/models/index.mjs';
import { setupAuthSecretSync } from './lib/system/index.mjs';


export async function main() {
    // Init Logger
    const logger = createLogger({
        context: {
            logLevel: ConfigContainer.config.logLevel,
            serviceName: ConfigContainer.config.serviceName,
        },
        type: LoggerTypes.Winston,
    });
    App.setLogger(logger);

    // Read secret file and add to the configuration
    setupAuthSecretSync(ConfigContainer.config.auth);

    // Init Repository Layer
    const repository = createRepository({
        mariadb: createMariaDbConnection({
            connectionsLimit: ConfigContainer.config.mariadb.connectionsLimit,
            database: ConfigContainer.config.mariadb.database,
            host: ConfigContainer.config.mariadb.host,
            idleTimeout: 0,
            password: ConfigContainer.config.mariadb.password,
            port: ConfigContainer.config.mariadb.port,
            user: ConfigContainer.config.mariadb.user,
        }),
        prefix: ConfigContainer.config.redis.prefix,
        redis: createRedisConnection({
            host: ConfigContainer.config.redis.host,
            port: ConfigContainer.config.redis.port,
        }),
    });

    // Init Domain Model Layer
    ModelBase.setRepository(repository);

    // Init context. Fill the context with the solution data
    if (process.env.NODE_ENV !== 'test' && process.env.NODE_ENV !== 'gitlab') {
        await State.init();
    }

    // Init Subscriptions State in Memory
    SubscriptionsManagerBase.init();

    // Init Controllers Layer (API)
    ZmqApi.startService({
        pubSubUrl: ConfigContainer.config.zmqPubSubUrl,
    });
    WsApi.startServer({
        wsPort: ConfigContainer.config.wsPort,
    });

    // Add Global Unhandled Errors Handlers
    async function exit() {
        await ZmqApi.stopServer();
        await WsApi.stopServer();
        logger.info('Exit');

        process.exit(0);
    }

    process.on('SIGTERM', async () => {
        logger.error('SIGTERM signal caught');
        await exit();
    });

    process.on('SIGINT', async () => {
        logger.error('SIGINT signal caught');
        await exit();
    });

    process.on('unhandledRejection', (error) => {
        logger.error('unhandledRejection', error.stack);
    });

    process.on('uncaughtException', (error) => {
        logger.error('uncaughtException', error.stack);
    });
}
