import { createRequire } from 'module';
import { randomInt } from '../utils/index.mjs';
import { generateUid } from '../../lib/utils/index.mjs';
import { main } from '../../appSetup.mjs';
import { Repository } from '../utils/Repository.mjs';

const require = createRequire(import.meta.url);
const DbSeeds = require('../seeds/data.json');

/**
 * @function
 * @param {String} mac - hardware device MAC address
 * @return {Object}
 */
export function createPosition(mac) {
    return {
        data: {
            er: null,
            fom: 50,
            inactive: true,
            mac,
            sl_uid: generateUid(),
            slvr_mode: 'UNKNOWN',
            type: 'UWB',
            x: randomInt(),
            y: randomInt(),
            z: randomInt(),
        },
        tk_uid: generateUid(),
        ts: new Date().toISOString(),
    };
}

/**
 * @function
 * @param {String} object - object name (ZONE, NODE, ...)
 * @param {String} action - action type (created, updated, ...)
 * @param {String} projectUid - project unique identifier
 * @return {Object}
 */
export function createNotify(object, action, projectUid) {
    const now = new Date().getTime();
    const ms = now % 1000;
    const ts = Math.floor((now - ms) / 1000);

    return {
        action,
        ms,
        object,
        plist: {},
        project_uid: projectUid,
        ts,
    };
}

/**
 * @function
 * @param {String} uid - alarm unique identifier
 * @param {String} projectUid - project unique identifier
 * @return {Object}
 */
export function createAlarm(uid, projectUid) {
    const now = new Date().getTime();
    const ms = now % 1000;
    const ts = Math.floor((now - ms) / 1000);

    return {
        alarm_name: 'zone_violation',
        created_at: new Date().toISOString(),
        ms,
        project_uid: projectUid,
        state: 'set',
        sublocation_uid: generateUid(),
        ts,
        uid,
        updated_at: new Date().toISOString(),
    };
}

/**
 * @function
 * @param {String} mac - hardware MAC address
 * @param {String} uid
 * @return {Object}
 */
export function createOta(mac, uid) {
    const now = new Date().getTime();
    const ms = now % 1000;
    const ts = Math.floor((now - ms) / 1000);

    return {
        id: randomInt(),
        mac,
        ms,
        plist: {
            crc: 'EFA6DB81',
            git: '5F013059',
            page_n: 84,
            page_sz: 512,
        },
        ts,
        type: 'ota_page_req',
        uid,
    };
}

/**
 * @function
 * @param {String} uid - alarm unique identifier
 * @param {String} projectUid - project unique identifier
 * @return {Object}
 */
export function createZoneEvent(uid, projectUid) {
    const now = new Date().getTime();
    const ms = now % 1000;
    const ts = Math.floor((now - ms) / 1000);

    return {
        action: 'leave',
        ms,
        project_uid: projectUid,
        state: 'set',
        sublocation_uid: generateUid(),
        ts,
        uid, // need as identificator for test
        x: randomInt(),
        y: randomInt(),
        z: randomInt(),
        zone_type_name: 'keep_in',
        zone_uid: generateUid(),
    };
}

let initialized;
export const Server = () => {
    const serverStart = () => (main());

    return {
        init: () => {
            if (!initialized) {
                initialized = serverStart();
            }
            return initialized;
        },
    };
};

export async function initRepoData() {
    const repository = new Repository();
    for (const migration of DbSeeds.migrations) {
        await repository.runSql(migration);
    }
    for (const user of DbSeeds.users) {
        await repository.createUser(user);
    }
    for (const node of DbSeeds.nodes) {
        await repository.createNode(node);
    }
}

export async function clearRepoData() {
    const repository = new Repository();

    for (const user of DbSeeds.users) {
        await repository.deleteUser(user.email);
    }
    for (const node of DbSeeds.nodes) {
        await repository.deleteNode(node.mac);
    }
}

export async function delay(ms) {
    return new Promise((resolve) => setTimeout(resolve, ms));
}
