import assert from 'assert';
import sinon from 'sinon';
import { ZmqPublisher } from '../utils/index.mjs';
import {
    clearRepoData, initRepoData, Server, delay,
} from './helpers.mjs';
import { Dispatcher, Notify } from '../../lib/models/index.mjs';
import { WS_CHANNELS, ZMQ_SUB_CHANNELS } from '../../lib/system/index.mjs';

const spy = {
    envelope: null,
    send: null,
};

const correctUserMsg = {
    company_uid: '2FGzD2v1SuKkueYdta2ppm',
    dev_id: 'dz3lkpN-RnmH7hqoftV-1Q',
    hex: 'CA20000E00000100000002000012000A0002',
    mac: '00000000001a',
    ms: 458,
    project_uid: 'wABnaPYiSLyHncJVmTWHE2',
    ts: 1733166298,
};

describe('ZMQ Interface', () => {
    const zmqPublisher = new ZmqPublisher();

    before(async () => {
        await Server().init();
        await zmqPublisher.init(); // init with some delay to init zmq connection
        await initRepoData();
    });

    after(async () => {
        await clearRepoData();
        zmqPublisher.destroy();
    });

    beforeEach(() => {
        spy.envelope = sinon.spy(Notify, 'envelope');
        spy.send = sinon.spy(Dispatcher, 'send');
    });

    afterEach(() => {
        spy.envelope.restore();
        spy.send.restore();
    });

    describe('USER_MSG', () => {
        it('1. should Dispatch correct send arguments incorrect mac', async () => {
            const userMsg = { ...correctUserMsg };
            const topic = `${WS_CHANNELS.USER_MSG}:${userMsg.project_uid}`;

            zmqPublisher.send(ZMQ_SUB_CHANNELS.USER_MSG, userMsg);
            await delay(50);

            assert(spy.envelope.withArgs(userMsg, WS_CHANNELS.USER_MSG).calledOnce);
            assert(spy.send.withArgs(topic, sinon.match.object).calledOnce);
            assert.deepEqual(userMsg, spy.send.getCall(0).args[1].params.data);
        });

        it('2. should Dispatch empty topic send on wrong mac', async () => {
            const userMsg = { ...correctUserMsg, mac: 'E0000000001a' };
            const topic = null;

            zmqPublisher.send(ZMQ_SUB_CHANNELS.USER_MSG, userMsg);
            await delay(50);

            assert.equal(spy.envelope.callCount, 0);
            assert(spy.send.withArgs(topic, sinon.match.object).calledOnce);
            assert.deepEqual(userMsg, spy.send.getCall(0).args[1]);
        });
    });
});
