import redis from 'redis';
import { promisify } from 'util';
import Redlock from 'redlock';
import RedisWildCard from 'redis-delete-wildcard';
import logger from '../api/logger.mjs';


RedisWildCard(redis);

/**
 * @function
 * @param {Object} options - database connection options
 * @param {String} options.host - database host name
 * @param {Number} options.port - database port number
 * @return {Object} - DB API
 */
export function createRedisConnection(options) {
    const REDLOCK_OFF = false;
    const ASSOC_LOCK_TTL = 5000;

    const client = redis.createClient(options);
    client.on('error', (error) => {
        console.error(error);
    });
    const redlock = new Redlock(
        // you should have one client for each independent redis node
        // or cluster
        [client],
        {
            // the expected clock drift; for more details
            // see http://redis.io/topics/distlock
            driftFactor: 0.01, // multiplied by lock ttl to determine drift time

            // the max number of times Redlock will attempt
            // to lock a resource before erroring
            retryCount: 10,

            // the time in ms between attempts
            retryDelay: 200, // time in ms

            // the max time in ms randomly added to retries
            // to improve performance under high contention
            // see https://www.awsarchitectureblog.com/2015/03/backoff.html
            retryJitter: 200, // time in ms
        },
    );

    /**
     * Execute inside lock
     * @param resourceKey {string} resource lock name
     * @param foo {Promise} to execute
     * @returns {Promise<void>}
     */
    async function executeWithLock(resourceKey, foo) {
        let result;
        let lock = null;
        try {
            lock = REDLOCK_OFF ? null : await redlock.lock(resourceKey, ASSOC_LOCK_TTL);
            result = await foo();
        } catch (e) {
            logger.error(`Error getting ${resourceKey} lock`, e);
            throw e;
        } finally {
            if (lock) lock.unlock();
        }
        return result;
    }

    redlock.on('clientError', (err) => {
        console.error('A redis error has occurred:', err);
    });
    return {
        client,
        decr: promisify(client.decr).bind(client),
        del: promisify(client.del).bind(client),
        delwild: promisify(client.delwild).bind(client),
        executeWithLock,
        get: promisify(client.get).bind(client),
        hdel: promisify(client.hdel).bind(client),
        hgetall: promisify(client.hgetall).bind(client),
        hmget: promisify(client.hmget).bind(client),
        hmset: promisify(client.hmset).bind(client),
        incr: promisify(client.incr).bind(client),
        lindex: promisify(client.lindex).bind(client),
        lpop: promisify(client.lpop).bind(client),
        lrange: promisify(client.lrange).bind(client),
        multi: promisify(client.multi).bind(client),
        redlock,
        rpush: promisify(client.rpush).bind(client),
        sadd: promisify(client.sadd).bind(client),
        scan: promisify(client.scan).bind(client),
        set: promisify(client.set).bind(client),
        smembers: promisify(client.smembers).bind(client),
        spop: promisify(client.spop).bind(client),
        srem: promisify(client.srem).bind(client),
        watch: promisify(client.watch).bind(client),
    };
}
