import { State } from '../state/index.mjs';
import { WS_CHANNELS, ZMQ_OBJECT } from '../../system/index.mjs';

export class Notify {
    /**
     * @class {Notify}
     * */

    static envelope(message, channel) {
        return {
            jsonrpc: '2.0',
            method: 'notify',
            params: {
                data: { ...message },
                subscription: channel,
            },
        };
    }

    /**
     * @method
     * @param {Object} positionRaw  - raw position form JSON/BIN convertor
     * @property {String} positionRaw.ts  - position timestamp in ISO 8601 format
     * @property {String} positionRaw.tk_uid  - trackable object uid
     * @property {Object} positionRaw.data  - position properties
     * @property {Number} positionRaw.data.x  - the X coordinate of the position
     * @property {Number} positionRaw.data.y  - the Y coordinate of the position
     * @property {Number} positionRaw.data.z  - the Z coordinate of the position
     * @property {String} positionRaw.data.type  - positions type (UWB, HB)
     * @property {String} positionRaw.data.sl_uid  - sublocation numeric identifier
     * @property {String} positionRaw.data.mac  - hardware MAC address
     * @property {String} positionRaw.data.fom  - the figure of merit
     * @property {String} positionRaw.data.er  - 0D mode - distance to 0D anchor in millimeters
     * @property {String} positionRaw.data.slvr_mode  - the PE solver mode (NONE, PD, 0D, 1D, TDOA, TWOA, UNKNOWN)
     * @property {String} positionRaw.data.inactive  - the flag that indicated the tag mode: false - tag is moving, true - tag is not moving
     * @returns {Promise<{topic: string|null, message: Object}>}
     */
    static async getPositionMessage(positionRaw) {
        const mac = (positionRaw.data.mac || '').toUpperCase();
        const macAssociation = await new State().readNodeProject(mac);
        if (!macAssociation) return { message: positionRaw, topic: null };

        const timestamp = new Date(positionRaw.ts).getTime();
        const ms = timestamp % 1000;
        const ts = (timestamp - ms) / 1000;
        const inactive = typeof positionRaw.data.inactive === 'boolean' ? positionRaw.data.inactive : null;

        const message = {
            company_uid: macAssociation.companyUid,
            dev_id: positionRaw.tk_uid || 'unassigned',
            edge_corr: null,
            er: positionRaw.data.er || null, // marked as deprecated in public API
            fom: positionRaw.data.fom || 0,
            id: null,
            inactive,
            locationId: null,
            mac, // marked as deprecated in public API
            ms, // marked as deprecated in public API
            pe_dim: null,
            project_uid: macAssociation.projectUid,
            sfsn: null,
            sl: null,
            sl_id: positionRaw.data.sl_uid || '', // marked as deprecated in public API
            slvr_mode: positionRaw.data.slvr_mode || null, // no value in a platform
            to_type_id: null,
            ts, // no value in a platform
            type: positionRaw.data.type,
            x: positionRaw.data.x, // no value in a platform
            y: positionRaw.data.y,
            z: positionRaw.data.z,
        };
        const topic = `${WS_CHANNELS.POS}:${macAssociation.projectUid}`;

        return {
            message: Notify.envelope(message, WS_CHANNELS.POS),
            topic,
        };
    }

    /**
     * Get Project uid from #notify message. Default value is in notifyRaw.project_uid
     * @param {Object} notifyRaw
     * @property {String} notifyRaw.object - ZMQ objects:['WIN'|'USER'|'PROJECT_USER_ACCESS',...]
     * @property {String} notifyRaw.action - ZMQ actions: ['created'|'updated'|'deleted'|'release'|'assign'].
     * @property {String} notifyRaw.mac - win mac_address used for object 'WIN'
     * @returns {Promise<string|*>}
     */
    static async getProjectUid(notifyRaw) {
        switch (notifyRaw.object) {
            case ZMQ_OBJECT.WIN: {
                const node = await (new State()).readNodeProject(notifyRaw.mac);
                return node && node.projectUid;
            }
            case ZMQ_OBJECT.USER:
            case ZMQ_OBJECT.PROJECT_USER_ACCESS:
            default: {
                return notifyRaw.project_uid;
            }
        }
    }

    /**
     * @method
     * @param {Object} notifyRaw - notify channel message
     * @property {Number} notifyRaw.ts - unix timestamp
     * @property {Number} notifyRaw.ms - milliseconds part of the timestamp
     * @property {String|undefined} notifyRaw.project_uid - project unique identifier
     * @property {String|undefined} notifyRaw.company_uid - company unique identifier
     * @property {String} notifyRaw.mac - device mac address
     * @property {String} notifyRaw.object - entity name
     * @property {String} notifyRaw.action - action name
     * @property {Object} notifyRaw.plist - payload
     * @return {Object<{topic: string, message: Object}>}
     */
    static async getNotifyMessage(notifyRaw) {
        const projectUid = await Notify.getProjectUid(notifyRaw);

        const topic = `${WS_CHANNELS.NOTIFY}:${projectUid}`;

        return {
            message: Notify.envelope(notifyRaw, WS_CHANNELS.NOTIFY),
            topic,
        };
    }

    /**
     * @method
     * @param {Object} alarmRaw - alarm channel message
     * @property {String} alarmRaw.project_uid - project unique identifier
     * @return {Object<{topic: string, message: Object}>}
     */
    static getAlarmMessage(alarmRaw) {
        const topic = `${WS_CHANNELS.ALARM}:${alarmRaw.project_uid}`;

        return {
            message: Notify.envelope(alarmRaw, WS_CHANNELS.ALARM),
            topic,
        };
    }

    /**
     * @method
     * @param {Object} sensorRawData - sensor channel message
     * @property {String} sensorRawData.company_uid - company unique identifier
     * @property {String} sensorRawData.project_uid - project unique identifier
     * @return {Object<{topic: string, message: Object}>}
     */
    static getSensorMessage(sensorRawData) {
        const topic = `${WS_CHANNELS.SENSOR}:${sensorRawData.project_uid}`;

        return {
            message: Notify.envelope(sensorRawData, WS_CHANNELS.SENSOR),
            topic,
        };
    }

    /**
     * @method
     * @param {Object} otaRawData - sensor channel message
     * @property {String} otaRawData.mac - hardware MAC address
     * @return {Promise<{topic: string|null, message: Object}>}
     */
    static async getOtaMessage(otaRawData) {
        const mac = (otaRawData.mac || '').toUpperCase();
        const macAssociation = await new State().readNodeProject(mac);
        if (!macAssociation) return { message: otaRawData, topic: null };
        const topic = `${WS_CHANNELS.OTA}:${macAssociation.projectUid}`;

        return {
            message: Notify.envelope(otaRawData, WS_CHANNELS.OTA),
            topic,
        };
    }

    /**
     * @method
     * @param {Object} userMsgRawData - sensor channel message
     * ....
     * @property {String} userMsgRawData.mac - hardware MAC address
     * @return {Promise<{topic: string|null, message: Object}>}
     */
    static async getUserMsgMessage(userMsgRawData) {
        const mac = (userMsgRawData.mac || '').toUpperCase();
        const macAssociation = await new State().readNodeProject(mac);
        if (!macAssociation) return { message: userMsgRawData, topic: null };
        const topic = `${WS_CHANNELS.USER_MSG}:${macAssociation.projectUid}`;

        return {
            message: Notify.envelope(userMsgRawData, WS_CHANNELS.USER_MSG),
            topic,
        };
    }

    /**
     * @method
     * @param {Object} zoneEventRaw - raw zone event
     * @property {String} zoneEventRaw.action - zone event action (enter, leave)
     * @property {Number} zoneEventRaw.ts - unix timestamp
     * @property {Number} zoneEventRaw.ms - millisecond part of te timestamp
     * @property {String} zoneEventRaw.obj_uid - trackable object uid
     * @property {String} zoneEventRaw.project_uid - project uid
     * @property {String} zoneEventRaw.sublocation_uid - sublocation uid
     * @property {Number} zoneEventRaw.x - zone event X coordinate
     * @property {Number} zoneEventRaw.y - zone event Y coordinate
     * @property {Number} zoneEventRaw.z - zone event Z coordinate
     * @property {String} zoneEventRaw.zone_type_name - zone type
     * @property {String} zoneEventRaw.zone_uid - zone unique identifier
     * @returns {Object<{topic: string, message: Object}>}
     */
    static getZoneEventMessage(zoneEventRaw) {
        const topic = `${WS_CHANNELS.ZONE_EVENT}:${zoneEventRaw.project_uid}`;

        return {
            message: Notify.envelope(zoneEventRaw, WS_CHANNELS.ZONE_EVENT),
            topic,
        };
    }

    /**
     * @method
     * @param {Object} paramPRaw - raw param_p message
     * @property {String} paramPRaw.mac - hardware MAC address
     * @property {Number} paramPRaw.ts - unix timestamp
     * @property {Number} paramPRaw.ms - milliseconds part of the timestamp
     * @property {Object} paramPRaw.plist - message payload
     * @returns {Promise<{topic: string|null, message: Object}>}
     */
    static async getParamPMessage(paramPRaw) {
        const macAssociation = await new State().readNodeProject(paramPRaw.mac);
        if (!macAssociation) return { message: paramPRaw, topic: null };
        const topic = `${WS_CHANNELS.PARAM_P}:${macAssociation.projectUid}`;

        return {
            message: Notify.envelope(paramPRaw, WS_CHANNELS.PARAM_P),
            topic,
        };
    }
}
