import assert from 'assert';
import { generateUid } from '../../utils/index.mjs';
import { Notify } from './Notify.mjs';
import ModelBase from '../ModelBase.mjs';

const projectUid = generateUid();
const companyUid = generateUid();
const repository = {
    state: {
        read: async function read(mac) {
            return { mac, projectUid, companyUid };
        },
        remove: async function remove(mac) {
            return mac;
        },
        // eslint-disable-next-line no-shadow
        save: async function save({ mac, projectUid, companyUid }) {
            return { mac, projectUid, companyUid };
        },
    },
};

describe('Notify', () => {
    before(() => {
        ModelBase.setRepository(repository);
    });

    it('should return position message', async () => {
        const trackableUid = generateUid();
        const ts = new Date().toISOString();
        const x = 1000;
        const y = 1000;
        const z = 1000;
        const type = 'UWB';
        const sublocationUid = generateUid();
        const mac = '00000000001a';
        const rawPosition = {
            data: {
                mac,
                sl_uid: sublocationUid,
                type,
                x,
                y,
                z,
            },
            tk_uid: trackableUid,
            ts,
        };
        const notify = await Notify.getPositionMessage(rawPosition);
        const { message: { params: { subscription, data }, method, jsonrpc } } = notify;
        assert(notify.topic === `pos:${projectUid}`, 'topic wrong value');
        assert(jsonrpc === '2.0', 'jsonrpc wrong value');
        assert(method === 'notify', 'notify wrong value');
        assert(subscription === 'pos', 'subscription wrong value');
        assert(new Date(data.ts * 1000 + data.ms).getTime() === new Date(ts).getTime(), 'timestamp wrong value');
        assert(data.x === x, 'x wrong value');
        assert(data.y === y, 'y wrong value');
        assert(data.z === z, 'z wrong value');
        assert(data.sl_id === sublocationUid, 'sl_id wrong value');
        assert(data.type === 'UWB', 'type wrong value');
        assert(data.mac === mac.toUpperCase(), 'mac wrong value');
        assert(data.dev_id === trackableUid, 'dev_id wrong value');
        assert(data.project_uid === projectUid, 'project_uid wrong value');
        assert(data.company_uid === companyUid, 'company_uid wrong value');
    });

    it('should return notify message', async () => {
        const now = new Date().getTime();
        const ms = now % 1000;
        const ts = Math.floor((now - ms) / 1000);
        const object = 'NODE';
        const action = 'updated';
        const plist = { uid: generateUid() };
        const rawNotify = {
            action,
            company_uid: companyUid,
            ms,
            object,
            plist,
            project_uid: projectUid,
            ts,
        };
        const notify = await Notify.getNotifyMessage(rawNotify);
        const { message: { params: { subscription, data }, method, jsonrpc } } = notify;
        assert(notify.topic === `notify:${projectUid}`, 'topic wrong value');
        assert(jsonrpc === '2.0', 'jsonrpc wrong value');
        assert(method === 'notify', 'notify wrong value');
        assert(subscription === 'notify', 'subscription wrong value');
        assert.deepStrictEqual(data, rawNotify);
    });

    it('should return alarm message', async () => {
        const now = new Date().getTime();
        const ms = now % 1000;
        const ts = Math.floor((now - ms) / 1000);
        const rawAlarm = {
            ms,
            plist: { uid: generateUid() },
            project_uid: projectUid,
            ts,
            type: 'zone_violation',
        };
        const notify = Notify.getAlarmMessage(rawAlarm);
        const { message: { params: { subscription, data }, method, jsonrpc } } = notify;
        assert(notify.topic === `alarm:${projectUid}`, 'topic wrong value');
        assert(jsonrpc === '2.0', 'jsonrpc wrong value');
        assert(method === 'notify', 'notify wrong value');
        assert(subscription === 'alarm', 'subscription wrong value');
        assert.deepStrictEqual(data, rawAlarm);
    });

    it('should return sensor message', async () => {
        const now = new Date().getTime();
        const ms = now % 1000;
        const ts = Math.floor((now - ms) / 1000);
        const rawSensor = {
            company_uid: companyUid,
            ms,
            plist: { uid: generateUid() },
            project_uid: projectUid,
            ts,
        };
        const notify = Notify.getSensorMessage(rawSensor);
        const { message: { params: { subscription, data }, method, jsonrpc } } = notify;
        assert(notify.topic === `sensor:${projectUid}`, 'topic wrong value');
        assert(jsonrpc === '2.0', 'jsonrpc wrong value');
        assert(method === 'notify', 'notify wrong value');
        assert(subscription === 'sensor', 'subscription wrong value');
        assert.deepStrictEqual(data, rawSensor);
    });

    it('should return ota message', async () => {
        const now = new Date().getTime();
        const ms = now % 1000;
        const ts = Math.floor((now - ms) / 1000);
        const mac = '00000000001a';
        const rawOta = {
            mac,
            ms,
            plist: {
                uid: generateUid(),
            },
            ts,
        };
        const notify = await Notify.getOtaMessage(rawOta);
        const { message: { params: { subscription, data }, method, jsonrpc } } = notify;
        assert(notify.topic === `ota:${projectUid}`, 'topic wrong value');
        assert(jsonrpc === '2.0', 'jsonrpc wrong value');
        assert(method === 'notify', 'notify wrong value');
        assert(subscription === 'ota', 'subscription wrong value');
        assert.deepStrictEqual(data, rawOta);
    });

    it('should return zone_event message', async () => {
        const now = new Date().getTime();
        const ms = now % 1000;
        const ts = Math.floor((now - ms) / 1000);
        const rawZoneEvent = {
            action: 'enter',
            ms,
            obj_uid: generateUid(),
            project_uid: projectUid,
            sublocation_uid: generateUid(),
            ts,
            x: 1000,
            y: 100,
            z: 1,
            zone_type_name: 'keep_in',
            zone_uid: generateUid(),
        };
        const notify = Notify.getZoneEventMessage(rawZoneEvent);
        const { message: { params: { subscription, data }, method, jsonrpc } } = notify;
        assert(notify.topic === `zone_event:${projectUid}`, 'topic wrong value');
        assert(jsonrpc === '2.0', 'jsonrpc wrong value');
        assert(method === 'notify', 'notify wrong value');
        assert(subscription === 'zone_event', 'subscription wrong value');
        assert.deepStrictEqual(data, rawZoneEvent);
    });

    it('should return param_p message', async () => {
        const now = new Date().getTime();
        const ms = now % 1000;
        const ts = Math.floor((now - ms) / 1000);
        const mac = '00000000001a';
        const rawParamP = {
            mac,
            ms,
            plist: { uid: generateUid() },
            ts,
        };
        const notify = await Notify.getParamPMessage(rawParamP);
        const { message: { params: { subscription, data }, method, jsonrpc } } = notify;
        assert(notify.topic === `param_p:${projectUid}`, 'topic wrong value');
        assert(jsonrpc === '2.0', 'jsonrpc wrong value');
        assert(method === 'notify', 'notify wrong value');
        assert(subscription === 'param_p', 'subscription wrong value');
        assert.deepStrictEqual(data, rawParamP);
    });
});
